from tatara_logging.span import Span
from tatara_logging._background_queue_logger import BackgroundLazyQueueLogger
from tatara.types import (
    ImageFormat,
    DiffusionParams,
    DiffusionPrompt,
    LLMPrompt,
    LLMParams,
    LLMUsageMetrics,
    LogType,
)
from tatara_logging.rating import Rating
from tatara_logging.utils import _check_event
from typing import Optional, Dict, Any
import uuid
import time
from typing import TYPE_CHECKING

# Imports during type checking only to avoid circular dependencies at runtime.
# _TraceImpl takes a _SpanImpl as an argument, and _SpanImpl takes a _TraceImpl, so
# we need to use TYPE_CHECKING to avoid circular dependencies.
if TYPE_CHECKING:
    from tatara_logging.trace_impl import _TraceImpl

from tatara_logging._record_keys import (
    LOG_FORMAT_VERSION,
    LOG_RECORD_KEY_EVENT,
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_INTERNAL_ID,
    LOG_RECORD_KEY_PARENT_ID,
    LOG_RECORD_KEY_PARENT_TRACE_ID,
    LOG_RECORD_KEY_PROJECT,
    LOG_RECORD_KEY_PROPERTIES,
    LOG_RECORD_KEY_TIMESTAMP,
    LOG_RECORD_KEY_TYPE,
    LOG_RECORD_KEY_VERSION,
    LOG_RECORD_PROPERTIES_KEY_END_TIME,
    LOG_RECORD_PROPERTIES_KEY_START_TIME,
    LOG_RECORD_PROPERTIES_KEY_LLM_EVENT,
    LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT,
    LOG_RECORD_KEY_METADATA,
)

from tatara.client_state import _get_client_state


class _SpanImpl(Span):
    def __init__(
        self,
        event: str,
        id_: str,
        parent_id: str,
        trace: "_TraceImpl",
        background_queue_logger: BackgroundLazyQueueLogger,
        start_time: Optional[float] = None,
    ):
        self._event = event
        self._internal_id = "s_" + str(uuid.uuid4())
        self._id = id_
        self._parent_id = parent_id
        self._trace = trace

        # self.parent_span_id = parent_span_id
        self._background_queue_logger = background_queue_logger
        now = time.time()
        self._start_time = start_time or now
        self._end_time = None
        self._properties = {
            LOG_RECORD_PROPERTIES_KEY_START_TIME: self._start_time,
            LOG_RECORD_PROPERTIES_KEY_END_TIME: None,
        }
        self._token = _get_client_state().current_span.set(self)

        self._metadata = {}

        self._log_record = {
            LOG_RECORD_KEY_PROJECT: _get_client_state().project,
            LOG_RECORD_KEY_TYPE: LogType.SPAN,
            LOG_RECORD_KEY_TIMESTAMP: now,
            LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
            LOG_RECORD_KEY_INTERNAL_ID: self._internal_id,
            LOG_RECORD_KEY_ID: self._id,
            LOG_RECORD_KEY_PARENT_TRACE_ID: self._trace.id,
            LOG_RECORD_KEY_PARENT_ID: self._parent_id,
            LOG_RECORD_KEY_EVENT: self._event,
            LOG_RECORD_KEY_PROPERTIES: self._properties,
            LOG_RECORD_KEY_METADATA: self._metadata,
        }

        self._background_queue_logger.log(self._log_record)

    @property
    def id(self) -> str:
        return self._id

    @property
    def trace_id(self):
        return self._trace.id

    @property
    def event(self):
        return self._event

    def _check_finished(self):
        if self._end_time:
            raise RuntimeError("Cannot call method on a finished trace.")

    def _log(self):
        self._log_record[LOG_RECORD_KEY_TIMESTAMP] = time.time()
        self._background_queue_logger.log(self._log_record)

    def log_llm_success(
        self,
        prompt: str | LLMPrompt,
        output: str,
        params: Optional[LLMParams] = None,
        usage_metrics: Optional[LLMUsageMetrics] = None,
    ):
        self._check_finished()

        llm_event = {
            "prompt": prompt,
            "output": output,
            "params": params,
            "usage_metrics": usage_metrics,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_LLM_EVENT] = llm_event
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_input(
        self,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ):
        diffusion_event = {
            "prompt": prompt,
            "params": params,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_success_with_image_url(
        self,
        image_url: str,
        prompt: str | DiffusionPrompt,
        params: Optional[DiffusionParams] = None,
    ):
        self._check_finished()

        diffusion_event = {
            "prompt": prompt,
            "image_url": image_url,
            "params": params,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_success_with_image_data(
        self,
        image_data: str,
        image_format: ImageFormat,
        prompt: str | DiffusionPrompt,
        params: Optional[DiffusionParams] = None,
    ):
        self._check_finished()

        diffusion_event = {
            "image_data": image_data,
            "image_format": image_format,
            "prompt": prompt,
            "params": params,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_rating(self, rating: Rating) -> None:
        return _get_client_state().log_rating(
            trace_id=self.trace_id, span_event=self.event, rating=rating
        )

    def log_metadata(self, metadata: Dict[str, Any]):
        self._check_finished()

        self._metadata.update(metadata)
        self._log_record[LOG_RECORD_KEY_METADATA] = self._metadata
        self._log()

    def end(self, end_time: Optional[float] = None) -> None:
        self._check_finished()
        _get_client_state().current_span.reset(self._token)

        self._end_time = end_time or time.time()

        self._log_record[LOG_RECORD_KEY_PROPERTIES][
            LOG_RECORD_PROPERTIES_KEY_END_TIME
        ] = self._end_time
        self._log()

    def start_span(self, event: str) -> Span:
        self._check_finished()
        _check_event(event)

        span_id = self._trace._generate_id_for_span_event_and_log(event)

        return _SpanImpl(
            event=event,
            id_=span_id,
            parent_id=self.id,
            trace=self._trace,
            background_queue_logger=self._background_queue_logger,
            start_time=time.time(),
        )

    def __enter__(self) -> Span:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback

        self.end()



import atexit
import json
from abc import ABC, abstractmethod

from evals.eval_types import EvalRun
from network._tatara_network_client import TataraNetworkClient

MIN_FLUSH_EVENTS = 100
MIN_FLUSH_SECONDS = 10


class RecorderBase(ABC):
    def __init__(self):
        self.results = []
        # flush_events results on exit
        atexit.register(self.flush_events)

    @abstractmethod
    def _flush_internal(self, results):
        pass

    @abstractmethod
    def flush_events(self):
        pass

    @abstractmethod
    def record_eval_row(self, eval_row):
        pass

    @abstractmethod
    def record_eval_run(self, eval_run):
        pass


class FileRecorder(RecorderBase):
    def __init__(self, event_filepath: str):
        super().__init__()
        self.event_filepath = event_filepath

    def record(self, eval_row: EvalRun):
        self.results.append(eval_row)
        if len(self.results) > MIN_FLUSH_EVENTS:
            self.flush_events()

    def flush_events(self):
        with open(self.event_filepath, "a") as f:
            for eval_row in self.results:
                f.write(json.dumps(eval_row.to_dict()) + "\n")
        self.results = []


class TataraRecorder(RecorderBase):
    def __init__(self, tatara_network_client: TataraNetworkClient):
        super().__init__()
        self.eval_rows = []
        self.tatara_network_client = tatara_network_client

    def record_eval_run(self, eval_run: EvalRun):
        self.tatara_network_client.send_eval_run_post_request(eval_run)



import requests
from requests.adapters import HTTPAdapter
from evals.eval_types import EvalRun
from typing import Optional, Dict, List, Any
import json
from network.endpoints import Endpoints
from requests.models import Response


class TataraNetworkClient:
    def __init__(
        self,
        api_key: Optional[str] = None,
        is_dev: bool = False,
    ):
        self._headers = {"Authorization": api_key, "Content-Type": "application/json"}
        self.endpoints = Endpoints(is_dev=is_dev)

        adapter = HTTPAdapter(max_retries=0)
        self._http = requests.Session()
        self._http.mount("https://", adapter)
        self._http.mount("http://", adapter)

    ### LOGS
    def send_logs_post_request(self, logs: str) -> Optional[Response]:
        return self.send_post_request(self.endpoints.log_endpoint, logs)

    ### DATASETS
    def send_create_dataset_post_request(self, dataset_name: str) -> Optional[Response]:
        return self.send_post_request(
            self.endpoints.dataset_endpoint, data=json.dumps(dataset_name)
        )

    def send_dataset_get_request(self, dataset_name: str) -> Response:
        return self.send_get_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}"
        )

    def send_insert_records_post_request(
        self, dataset_name: str, records: List[Dict[str, Any]]
    ) -> Optional[Response]:
        return self.send_post_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}/insert_records",
            data=json.dumps(records),
        )

    def send_attach_records_post_request(
        self, dataset_name: str, record_ids: List[str]
    ) -> Optional[Response]:
        return self.send_post_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}/attach_records",
            data=json.dumps(record_ids),
        )

    ### EVALS
    def send_eval_run_post_request(self, eval_run: EvalRun) -> Optional[Response]:
        return self.send_post_request(
            f"{self.endpoints.eval_endpoint}/write", data=json.dumps(eval_run.to_dict())
        )

    ### GENERIC REQUESTS
    def send_post_request(self, endpoint: str, data: str) -> Optional[Response]:
        response = self._http.request(
            "POST",
            endpoint,
            data=data,
            headers=self._headers,
        )
        response.raise_for_status()
        return response

    def send_get_request(
        self, endpoint: str, params: Optional[dict] = None
    ) -> Response:
        response = self._http.get(
            endpoint,
            headers=self._headers,
            params=params,
        )
        response.raise_for_status()
        return response



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional

from evals.eval_types import EvalRecord, EvalRow, EvalRun, EvalValue
from evals.model_package import ModelInputType, ModelOutputType
from evals.record import Record
from evals.recorder import RecorderBase


class Eval(ABC):
    name: str
    description: str
    valid_input_types: List[ModelInputType]
    valid_output_types: List[ModelOutputType]

    def __init__(
        self,
        records: list,
        seed: Optional[int] = 42,
    ):
        self.seed = seed
        self.records = records

    @abstractmethod
    def eval_record(self, record: Record) -> Optional[bool]:
        """
        Eval a single record
        """
        raise NotImplementedError

    def run_no_recorder(self) -> List[EvalValue]:
        """
        Run the eval without recording the results anywhere
        """
        eval_results = []
        for record in self.records:
            eval_result = EvalValue(
                record_id=record.record_id,
                input=record.input,
                output=record.output,
                eval_record=EvalRecord(
                    eval_name=self.name,
                    eval_description=self.description,
                    result=self.eval_record(record),
                ),
            )
            eval_results.append(eval_result)

        return eval_results

    def run(self, recorder: RecorderBase) -> None:
        """
        Run the eval with the RecorderBase
        """
        for record in self.records:
            eval_row = {
                "record_id": record.record_id,
                "input": record.input,
                "output": record.output,
                "eval": {
                    "name": self.name,
                    "description": self.description,
                    "result": self.eval_record(record),
                },
            }
            recorder.record_eval_row(eval_row)

    def is_valid_input_type(self, model_input_type: ModelInputType) -> bool:
        return model_input_type in self.valid_input_types

    def is_valid_output_type(self, model_output_type: ModelOutputType) -> bool:
        logging.warning("")
        return model_output_type in self.valid_output_types


@dataclass
class Evals:
    evals: List[Eval]

    def eval_values_to_rows(
        self, all_eval_results: List[List[EvalValue]]
    ) -> List[EvalRow]:
        eval_rows = {}
        for eval_results in all_eval_results:
            for eval_result in eval_results:
                # join the eval result for the record into a single row
                record_id = eval_result.record_id
                if record_id in eval_rows:
                    # just add the eval result to the existing record in the output
                    eval_rows[record_id].add(eval_result)
                else:
                    eval_rows[record_id] = EvalRow(
                        record_id=record_id,
                        input=eval_result.input,
                        output=eval_result.output,
                        eval_values=[eval_result.eval_record],
                    )
        return list(eval_rows.values())

    def run(self, recorder: RecorderBase):
        all_eval_results = []
        for eval in self.evals:
            eval_results: List[EvalValue] = eval.run_no_recorder()
            all_eval_results.append(eval_results)

        eval_rows = self.eval_values_to_rows(all_eval_results)
        eval_run = EvalRun(eval_rows=eval_rows)
        # TODO: record the eval run as well via a separate endpoint in the recorder
        recorder.record_eval_run(eval_run)



from abc import ABC, abstractmethod
from tatara_logging.rating import Rating
from typing import Any, Dict, List, Optional
from tatara_logging.span import Span


class Trace(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """ID of the trace. If the user does not provide an ID, equivalent to trace_id"""

    @property
    @abstractmethod
    def event(self) -> str:
        """Caller provided event name of the trace"""

    @property
    @abstractmethod
    def span_ids(self) -> List[str]:
        """List of spans associated with the trace"""

    @abstractmethod
    def start_span(self, event: str, parent_event: Optional[str]) -> Span:
        """Start a span. Optionally provide a parent span event. If none is provided, the current trace is used as the parent."""

    @abstractmethod
    def end(self) -> None:
        """End the trace"""

    @abstractmethod
    def log_metadata(self, metadata: Dict[str, Any]) -> None:
        """Log arbitrary metadata to the trace"""

    @abstractmethod
    def log_rating(self, rating: Rating) -> None:
        """Log a rating to the trace"""

    @abstractmethod
    def __enter__(self) -> "Trace":
        """Enter context manager"""

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit context manager"""



import time
from inspect import iscoroutinefunction
from typing import Optional
import wrapt
import logging
from tatara_logging.trace import Trace
from tatara_logging.span import Span
from tatara_logging.empty_span import _EmptySpan
from tatara_logging.empty_trace import _EmptyTrace
from tatara_logging.trace_impl import _TraceImpl
from tatara_logging.rating import Rating
from tatara.types import DiffusionPrompt, DiffusionParams, LogType, ImageFormat
from tatara_logging.utils import _gen_id_from_trace_and_event
from tatara_logging._record_keys import (
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_PROJECT,
    LOG_RECORD_KEY_PROPERTIES,
    LOG_RECORD_KEY_TIMESTAMP,
    LOG_RECORD_KEY_TYPE,
    LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT,
)
import os

from client_state import TataraClientState, _get_client_state


def init(project: str, api_key: Optional[str] = None):
    if api_key is None:
        if os.environ.get("TATARA_API_KEY") is not None:
            api_key = os.environ.get("TATARA_API_KEY")
        else:
            raise ValueError("TATARA_API_KEY environment variable must be set.")

    global _tatara_client_state
    _tatara_client_state = TataraClientState(
        project,
        api_key,
    )


def current_trace() -> Trace:
    """
    Get the current trace. If there is no current trace, returns an empty trace.
    """
    current_trace = _get_client_state().current_trace.get()

    if current_trace is None:
        logging.log(
            logging.WARN, "No current trace found. Did you forget to start the trace?"
        )
        return _EmptyTrace()

    return current_trace


def current_span() -> Span:
    """
    Get the current trace. If there is no current trace, returns an empty trace.
    """
    current_span = _get_client_state().current_span.get()

    if current_span is None:
        logging.log(
            logging.WARN, "No current span found. Did you forget to start the span?"
        )
        return _EmptySpan()

    return current_span


def rate_trace(trace_id: str, rating: Rating):
    """
    Rate a trace.
    """
    # TODO: trace rating
    pass


def start_trace(
    event: str, id_: Optional[str] = None, user_id: Optional[str] = None
) -> Trace:
    """
    Create a new trace.
    """
    return _TraceImpl(
        event, logger=_get_client_state().bglq_logger, id_=id_, user_id=user_id
    )


def start_span(event: str, parent_event: Optional[str] = None) -> Span:
    """
    Create a new span. Automatically finds the current trace and returns an empty span an active trace is not found.
    """
    current_trace = _get_client_state().current_trace.get()

    if current_trace is None:
        return _EmptySpan()

    return current_trace.start_span(event, parent_event)


def log_diffusion_input(
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    log = {
        LOG_RECORD_KEY_ID: _gen_id_from_trace_and_event(trace_id, event),
        LOG_RECORD_KEY_PROJECT: _get_client_state().project,
        LOG_RECORD_KEY_TYPE: LogType.SPAN,
        LOG_RECORD_KEY_TIMESTAMP: time.time(),
        LOG_RECORD_KEY_PROPERTIES: {
            LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT: {
                "prompt": prompt,
                "params": params,
            }
        },
    }
    return _get_client_state().bglq_logger.log(log)


def log_diffusion_output_with_image_data(
    trace_id: str, event: str, image_data: str, image_format: ImageFormat
):
    pass


def log_diffusion_output_with_image_url(image_url: str, trace_id: str, event: str):
    log = {
        LOG_RECORD_KEY_ID: _gen_id_from_trace_and_event(trace_id, event),
        LOG_RECORD_KEY_PROJECT: _get_client_state().project,
        LOG_RECORD_KEY_TYPE: LogType.SPAN,
        LOG_RECORD_KEY_TIMESTAMP: time.time(),
        LOG_RECORD_KEY_PROPERTIES: {
            LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT: {
                "image_url": image_url,
            }
        },
    }
    return _get_client_state().bglq_logger.log(log)


def log_diffusion_call_with_image_data(
    image_data: str,
    image_format: ImageFormat,
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    pass


def log_diffusion_call_with_image_url(
    image_url: str,
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    pass


def log_rating_for_span(span_event: str, trace_id: str, rating: Rating):
    """
    Add a rating for a span.
    """
    _get_client_state().log_rating(
        rating=rating, trace_id=trace_id, span_event=span_event
    )


def log_rating_for_trace(id: str, rating: Rating):
    """
    Add a rating for a trace. ID can be either the user-provided id or the trace_id.
    """
    _get_client_state().log_rating(rating=rating, trace_id=id, span_event=None)


def log_trace(event: str):
    if callable(event):
        raise TypeError(
            "The 'log_trace' decorator expects an 'event' argument. Usage: @log_trace(event='your_event')"
        )

    def wrapper(wrapped):
        @wrapt.decorator
        async def _async_trace_wrapper(wrapped, instance, args, kwargs):
            trace_id = kwargs.pop("trace_id", None)
            user_id = kwargs.pop("tatara_user_id", None)

            with start_trace(event=event, id_=trace_id, user_id=user_id):
                return await wrapped(*args, **kwargs)

        @wrapt.decorator
        def _trace_wrapper(wrapped, instance, args, kwargs):
            trace_id = kwargs.pop("trace_id", None)
            user_id = kwargs.pop("tatara_user_id", None)

            with start_trace(event=event, id_=trace_id, user_id=user_id):
                return wrapped(*args, **kwargs)

        if iscoroutinefunction(wrapped):
            return _async_trace_wrapper(wrapped)  # type: ignore
        else:
            return _trace_wrapper(wrapped)  # type: ignore

    return wrapper


def log_span(event: Optional[str] = None, parent_event: Optional[str] = None):
    if callable(event):
        # Decorator was used without arguments, and 'event' is the function to be decorated.
        func = event
        return _create_span_decorator(event=None, parent_event=None)(func)
    else:
        # Decorator was used with string arguments.
        def decorator(func):
            return _create_span_decorator(event=event, parent_event=parent_event)(func)

        return decorator


def _create_span_decorator(event: Optional[str], parent_event: Optional[str] = None):
    def _check_event(event: Optional[str]):
        if event is None:
            raise TypeError(
                """Missing required 'event' argument. The 'log_span' decorator expects either:

1) An 'event' argument directly passed in to the decorator: 

    @log_span(event='your_event')
    def decorated_function():
        # do stuff

2) A 'tatara_event' argument passed from the caller of the decorated function:

    @log_span
    def decorated_function():
        # do stuff
    
    decorated_function(tatara_event='your_event') 
"""
            )

    def wrapper(wrapped):
        @wrapt.decorator
        async def _async_span_wrapper(wrapped, instance, args, kwargs):
            _event = kwargs.pop("tatara_event", event)
            _check_event(_event)
            _parent_event = kwargs.pop("tatara_parent_event", parent_event)
            with start_span(_event, _parent_event) as span:  # noqa: F841
                return await wrapped(*args, **kwargs)

        @wrapt.decorator
        def _span_wrapper(wrapped, instance, args, kwargs):
            _event = kwargs.pop("tatara_event", event)
            _check_event(_event)
            _parent_event = kwargs.pop("tatara_parent_event", parent_event)
            with start_span(_event, _parent_event) as span:  # noqa: F841
                return wrapped(*args, **kwargs)

        if iscoroutinefunction(wrapped):
            return _async_span_wrapper(wrapped)  # type: ignore
        else:
            return _span_wrapper(wrapped)  # type: ignore

    return wrapper



import hashlib
import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Union


class ModelInputType(Enum):
    TEXT = "text"
    IMAGE = "image"


class ModelOutputType(Enum):
    TEXT = "text"
    IMAGE = "image"


@dataclass
class ModelPackage:
    model_name: str
    hyperparameters: Dict[
        str, Union[str, float, int]
    ]  # this includes all of the hyperparameters for the model including the system prompt if there is one
    model_input_type: ModelInputType
    model_output_type: ModelOutputType
    model_sha256: str = field(
        init=False
    )  # a sha256 that uniquely identifies the model and hyperparamters

    @staticmethod
    def compute_model_sha256(
        model_name: str, hyperparameters: Dict[str, Union[str, float, int]]
    ) -> str:
        sorted_hyperparameters: str = json.dumps(hyperparameters, sort_keys=True)
        model_hash = hashlib.sha256(
            (model_name + sorted_hyperparameters).encode()
        ).hexdigest()
        return f"modelsha256_{model_hash}"

    def __post_init__(self):
        self.model_hash = ModelPackage.compute_model_sha256(
            self.model_name, self.hyperparameters
        )

    def __str__(self):
        return f"ModelPackage(model_name={self.model_name}, model_hash={self.model_hash[:20]}..., input_type={self.model_input_type}, output_type={self.model_output_type})"

    def __repr__(self):
        return self.__str__()

    def to_dict(self):
        return {
            "model_name": self.model_name,
            "hyperparameters": self.hyperparameters,
            "model_input_type": self.model_input_type.value,
            "model_output_type": self.model_output_type.value,
            "model_sha256": self.model_sha256,
        }



from dataclasses import dataclass, field
from typing import Any, Dict, List

from evals.id_generator import IdGenerator

RecordId = str


@dataclass
class EvalRecord:
    eval_name: str
    eval_description: str
    result: bool | None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "eval_name": self.eval_name,
            "eval_description": self.eval_description,
            "result": self.result,
        }


@dataclass
class EvalValue:
    record_id: RecordId
    input: str
    output: str
    eval_record: EvalRecord

    def to_dict(self) -> Dict[str, Any]:
        return {
            "record_id": self.record_id,
            "input": self.input,
            "output": self.output,
            "eval_record": self.eval_record.__dict__,
        }


@dataclass
class EvalRow:
    id: str = field(init=False)
    record_id: RecordId
    input: str
    output: str
    eval_values: List[EvalRecord]

    def __post_init__(self):
        self.id = IdGenerator.generate_eval_run_row_uuid()

    def add(self, eval_value: EvalValue) -> None:
        self.eval_values.append(eval_value.eval_record)

    def __iter__(self):
        return iter(self.eval_values)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "record_id": self.record_id,
            "input": self.input,
            "output": self.output,
            "eval_values": [eval_value.__dict__ for eval_value in self.eval_values],
        }


@dataclass
class EvalRun:
    id: str = field(init=False)
    eval_rows: List[EvalRow]

    def __post_init__(self):
        self.id = IdGenerator.generate_eval_run_uuid()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "eval_rows": [eval_row.to_dict() for eval_row in self.eval_rows],
        }



from tatara_logging.trace import Trace
from tatara_logging._background_queue_logger import BackgroundLazyQueueLogger
from tatara.types import (
    LogType,
)
from tatara_logging.span_impl import _SpanImpl
from tatara_logging.rating import Rating
from tatara_logging.utils import _check_event
from typing import Optional, Dict, Any, List
import uuid
import time

from tatara_logging._record_keys import (
    LOG_FORMAT_VERSION,
    LOG_RECORD_KEY_EVENT,
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_INTERNAL_ID,
    LOG_RECORD_KEY_PROJECT,
    LOG_RECORD_KEY_PROPERTIES,
    LOG_RECORD_KEY_TIMESTAMP,
    LOG_RECORD_KEY_TYPE,
    LOG_RECORD_KEY_VERSION,
    LOG_RECORD_PROPERTIES_KEY_END_TIME,
    LOG_RECORD_PROPERTIES_KEY_START_TIME,
    LOG_RECORD_KEY_METADATA,
    LOG_RECORD_KEY_SPANS,
    LOG_RECORD_KEY_USER_ID,
)

from tatara.client_state import _get_client_state


class _TraceImpl(Trace):
    def __init__(
        self,
        event,
        logger: BackgroundLazyQueueLogger,
        id_: Optional[str] = None,
        user_id: Optional[str] = None,
        start_time: Optional[float] = None,
    ):
        _check_event(event)
        self._event = event
        self._internal_id = "t_" + str(uuid.uuid4())
        self._id = id_ or self._internal_id
        self._user_id = user_id
        self._logger = logger
        now = time.time()
        self._start_time = start_time or now
        self._end_time = None
        self._span_ids = []
        self._metadata = {}
        self._token = _get_client_state().current_trace.set(self)

        self._properties = {
            LOG_RECORD_PROPERTIES_KEY_START_TIME: self._start_time,
            LOG_RECORD_PROPERTIES_KEY_END_TIME: None,
        }

        self._log_record = {
            LOG_RECORD_KEY_PROJECT: _get_client_state().project,
            LOG_RECORD_KEY_TYPE: LogType.TRACE,
            LOG_RECORD_KEY_TIMESTAMP: now,
            LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
            LOG_RECORD_KEY_INTERNAL_ID: self._internal_id,
            LOG_RECORD_KEY_ID: self._id,
            LOG_RECORD_KEY_USER_ID: self._user_id,
            LOG_RECORD_KEY_EVENT: self._event,
            LOG_RECORD_KEY_SPANS: self._span_ids,
            LOG_RECORD_KEY_PROPERTIES: self._properties,
            LOG_RECORD_KEY_METADATA: self._metadata,
        }

        self._logger.log(self._log_record)

    @property
    def id(self) -> str:
        return self._id

    @property
    def event(self) -> str:
        return self._event

    @property
    def span_ids(self) -> List[str]:
        return self._span_ids

    def _generate_id_for_span_event_and_log(self, event: str) -> str:
        span_id = event + "_" + self.id

        # add suffix if we're calling the same span event multiple times
        suffix = 2
        while span_id in self.span_ids:
            span_id = event + "_" + str(suffix) + "_" + self.id
            suffix += 1

        self._span_ids.append(span_id)

        self._log_record[LOG_RECORD_KEY_SPANS] = self._span_ids
        self._log()

        return span_id

    def start_span(self, event: str, parent_event: Optional[str]):
        self._check_finished()
        _check_event(event)

        span_id = self._generate_id_for_span_event_and_log(event)
        # TODO: handle edge case if multiple spans with the same event are called
        parent_id = (
            self.id
            if (parent_event is None or parent_event == self.event)
            else parent_event + "_" + self.id
        )
        return _SpanImpl(
            event=event,
            id_=span_id,
            parent_id=parent_id,
            trace=self,
            background_queue_logger=_get_client_state().bglq_logger,
            start_time=time.time(),
        )

    def end(self, end_time: Optional[float] = None):
        self._check_finished()

        self._end_time = end_time or time.time()
        _get_client_state().current_trace.reset(self._token)

        self._log_record[LOG_RECORD_KEY_PROPERTIES][
            LOG_RECORD_PROPERTIES_KEY_END_TIME
        ] = self._end_time
        self._log()

    def log_metadata(self, metadata: Dict[str, Any]):
        self._check_finished()

        self._metadata.update(metadata)
        self._log_record[LOG_RECORD_KEY_METADATA] = self._metadata
        self._log()

    # we can log a rating after a trace is finished
    def log_rating(self, rating: Rating):
        return _get_client_state().log_rating(
            rating=rating, trace_id=self.id, span_event=None
        )

    def _check_finished(self):
        if self._end_time:
            raise RuntimeError("Cannot call method on a finished trace.")

    def _log(self):
        self._log_record[LOG_RECORD_KEY_TIMESTAMP] = time.time()
        self._logger.log(self._log_record)

    def __enter__(self) -> Trace:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback

        self.end()



from typing import Optional
import re


def _gen_id_from_trace_and_event(trace_id: str, event: Optional[str]) -> str:
    if event:
        return event + "_" + trace_id
    return trace_id


def _check_event(event: str) -> None:
    if len(event) > 64:
        raise ValueError(
            f"'{event}' is too long of a event. Must be under 64 characters."
        )
    if not re.match(r"^\w+$", event):
        raise ValueError(
            f"'{event}' is an invalid event. Must consist of only alphanumeric characters and underscores."
        )



from dataclasses import dataclass
from typing import Optional, Dict
from provider_enum import ProviderEnum
from enum import Enum


@dataclass
class LLMUsageMetrics:
    prompt_tokens: int
    completion_tokens: int


@dataclass
class LLMPrompt:
    system_prompt: Optional[str] = None
    prompt_template: Optional[str] = None
    input_variables: Optional[Dict[str, str]] = None


@dataclass
class LLMParams:
    frequency_penalty: Optional[float] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    model: Optional[str] = None
    provider: Optional[str | ProviderEnum] = None


@dataclass
class DiffusionPrompt:
    prompt_template: Optional[str] = None
    input_variables: Optional[Dict[str, str]] = None
    negative_prompt: Optional[str] = None


@dataclass
class DiffusionParams:
    steps: Optional[int] = None
    height: Optional[int] = None
    width: Optional[int] = None
    model: Optional[str] = None
    provider: Optional[str | ProviderEnum] = None


class LogType(Enum):
    TRACE = 0
    SPAN = 1
    RATING = 2
    EVAL = 3


class ImageFormat(Enum):
    PNG = "png"
    JPG = "jpg"



import logging
from typing import Optional, Dict, List, Any
from evals.id_generator import IdGenerator
from dataclasses import dataclass
from requests.models import Response
from requests.exceptions import HTTPError
from client_state import _get_network_client


@dataclass
class Dataset:
    name: str
    records: List[Dict[str, Any]]
    """
    You should never instantiate a Dataset directly. Instead, use the `init_dataset` function.
    """

    def __iter__(self):
        return iter(self.records)

    def __str__(self) -> str:
        return f"Dataset(name={self.name})\n\n{self.records.__str__()}"

    def __repr__(self) -> str:
        return f"Dataset(name={self.name})\n\n{self.records.__repr__()}"

    @property
    def size(self) -> int:
        return len(self.records)

    def _is_valid_record(self, record: Dict[str, Any]) -> bool:
        required_fields = {"input", "output"}
        for required_field in required_fields:
            if required_field not in record.keys():
                return False
        return True

    def insert(self, records: List[Dict[str, Any]]) -> Optional[Dict]:
        """
        Insert a record to a dataset that consists of input, output, and metadata
        """
        for record in records:
            if not self._is_valid_record(record):
                logging.warning(
                    f"Record {record} is not valid. must have input and output fields"
                )

        resp = _get_network_client().send_insert_records_post_request(
            dataset_name=self.name, records=records
        )
        if resp and resp.ok:
            # Add the records to the current dataset
            self.records.extend(records)

    def attach_records(self, record_ids: List[str]) -> None:
        # attach records to a dataset that already exist
        IdGenerator.validate_record_ids(record_ids)
        resp = _get_network_client().send_attach_records_post_request(
            dataset_name=self.name, record_ids=record_ids
        )
        if resp and resp.ok:
            # Add the records to the current dataset
            records = resp.json()
            self.records.extend(records)

    def attach_record(self, record_id: str) -> None:
        # attach a record to a dataset that already exists
        self.attach_records([record_id])


def init_dataset(name: str) -> Dataset:
    """
    Initialize a dataset. This will create a dataset on the tatara server asynchonously
    and return a new dataset object
    """
    _get_network_client().send_create_dataset_post_request(name)
    return Dataset(name=name, records=[])


def get_dataset(name: str) -> Dataset:  # type: ignore
    """
    Get a dataset from the tatara server by name
    """
    try:
        ds_data_response: Response = _get_network_client().send_dataset_get_request(
            name
        )
        ds_data = ds_data_response.json()
        return Dataset(name=name, records=ds_data["records"])
    except HTTPError as e:
        if e.response.status_code == 404:
            print(f'`get_dataset` failed. The dataset "{name}" does not exist.')
        else:
            raise
    except Exception as e:
        print(f"Exception occurred when getting dataset: {e}")



from typing import Optional
import time
import contextvars
from tatara_logging.utils import _gen_id_from_trace_and_event

from tatara.types import LogType
from tatara_logging._record_keys import (
    LOG_FORMAT_VERSION,
    LOG_RECORD_KEY_HAS_RATING,
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_PROJECT,
    LOG_RECORD_KEY_RATING,
    LOG_RECORD_KEY_TIMESTAMP,
    LOG_RECORD_KEY_TYPE,
    LOG_RECORD_KEY_VERSION,
)
from tatara_logging.trace import Trace
from tatara_logging.span import Span
from tatara_logging.rating import Rating
from tatara_logging._background_queue_logger import BackgroundLazyQueueLogger
from network._tatara_network_client import TataraNetworkClient

DEFAULT_QUEUE_SIZE = 1000
DEFAULT_FLUSH_INTERVAL = 60.0

_tatara_client_state = None


class TataraClientState:
    current_trace: contextvars.ContextVar[Optional[Trace]]
    current_span: contextvars.ContextVar[Optional[Span]]

    def __init__(
        self,
        project: str,
        api_key: Optional[str],
        is_dev: bool = False,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        flush_interval: float = DEFAULT_FLUSH_INTERVAL,
    ):
        self.project = project
        self.api_key = api_key
        self.is_dev = is_dev
        self.current_trace = contextvars.ContextVar("current_trace", default=None)
        self.current_span = contextvars.ContextVar("current_span", default=None)

        self.tatara_network_client = TataraNetworkClient(
            api_key=self.api_key, is_dev=is_dev
        )
        self.bglq_logger = BackgroundLazyQueueLogger(
            queue_size, flush_interval=flush_interval, api_key=api_key
        )

    def log_rating(
        self, rating: Rating, trace_id: str, span_event: Optional[str] = None
    ):
        id = _gen_id_from_trace_and_event(trace_id, span_event)
        now = time.time()

        self.bglq_logger.log(
            {
                LOG_RECORD_KEY_PROJECT: self.project,
                LOG_RECORD_KEY_TYPE: LogType.TRACE
                if span_event is None
                else LogType.SPAN,
                LOG_RECORD_KEY_ID: id,
                LOG_RECORD_KEY_TIMESTAMP: now,
                LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
                LOG_RECORD_KEY_RATING: rating,
                LOG_RECORD_KEY_HAS_RATING: 1,
            }
        )

        if span_event is not None:
            self.bglq_logger.log(
                {
                    LOG_RECORD_KEY_PROJECT: self.project,
                    LOG_RECORD_KEY_TYPE: LogType.TRACE,
                    LOG_RECORD_KEY_ID: id,
                    LOG_RECORD_KEY_TIMESTAMP: now,
                    LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
                    LOG_RECORD_KEY_HAS_RATING: 1,
                }
            )


def _get_client_state() -> TataraClientState:
    if _tatara_client_state is None:
        raise Exception(
            "Tatara Client State not initialized. Please call init() before using the client."
        )
    return _tatara_client_state


def _get_network_client() -> TataraNetworkClient:
    if _tatara_client_state is None:
        raise Exception(
            "Tatara Client State not initialized. Please call init() before using the client."
        )
    return _tatara_client_state.tatara_network_client



from enum import Enum


class ProviderEnum(Enum):
    ANTHROPIC = "anthropic"
    ANYSCALE = "anyscale"
    FIREWORKS = "fireworks"
    HUGGING_FACE = "huggingface"
    OPENAI = "openai"
    STABILITYAI = "stabilityai"
    TOGETHERAI = "togetherai"
    OTHER = "other"



from dataclasses import dataclass

TATARA_SERVER_URL = "https://evals-fastapi.onrender.com"
LOCAL_SERVER_URL = "http://localhost:8080"


@dataclass
class Endpoint:
    server_url: str
    path: str

    def __post_init__(self):
        # Ensure that the path starts with a /
        if not self.path.startswith("/"):
            raise ValueError("Path must start with a /")

    @property
    def url(self) -> str:
        return f"{self.server_url}{self.path}"


class Endpoints:
    def __init__(self, is_dev: bool):
        self.server_url = LOCAL_SERVER_URL if is_dev else TATARA_SERVER_URL

    def _create_endpoint(self, path: str) -> str:
        return Endpoint(
            server_url=self.server_url,
            path=path,
        ).url

    @property
    def log_endpoint(self) -> str:
        return self._create_endpoint("/log")

    @property
    def dataset_endpoint(self) -> str:
        return self._create_endpoint("/dataset")

    @property
    def eval_endpoint(self) -> str:
        return self._create_endpoint("/eval")



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class BinaryRating(Enum):
    UPVOTE = 1
    NOVOTE = 0
    DOWNVOTE = -1


@dataclass
class Rating:
    rating: BinaryRating  # TODO: support other rating types (i.e. star scale, slider)
    feedback: Optional[str] = None
